;;; INTERPRETER definition

(require scheme/serialize)

(load "env.scm")
(load "structures.scm")
(load "parser.scm")
(load "dummy-transport-layer.scm")
(load "server.scm")
(load "prelude.scm")

(define no-host '((name . #f) (port . #f)))

;; parses and interprets the given program.
(define (run pgm)
  ;; interpreter in a non-server mode
  (register-host-properties (current-thread) no-host)
  (let ((init '(define current-host (quote ((name . #f) (port . #f))))))
    (run-init init pgm)))

(define (run-init init pgm)
  (let ((root-env (make-empty-env))
        (root-jp #f))
    
    (register-root-env (current-thread) root-env)
    (register-handler-thread (current-thread) (current-thread))
    (interp (parse (wrap init pgm)) root-env '() root-jp)))

(define (wrap init pgm)
  (let ((prog `(begin ,init ,*prelude* ,pgm))) ;; add the prelude and given init code
  ; (printf "~a~n" prog)
    prog))
 
;; evaluate expression exp in lexical environment E and aspect environment A, with current join point jp
(define (interp exp E A jp)
  (cond 
    
    ((def-exp? exp)
     (let ((val (interp (def-exp-exp exp) E A jp)))
       (augment-env (def-exp-id exp) val E)
       'void))
    
    ((lit-exp? exp) (lit-exp-value exp))
    
    ((prim-exp? exp)
     (let ((args (interp-args (prim-exp-args exp) E A jp)))
       (apply (eval (prim-exp-prim exp)) args)))
    
    ((and-exp? exp)
     (let loop ((args (and-exp-args exp)))
       (if (null? args)
           #t
           (and (interp (car args) E A jp)
                (loop (cdr args))))))
    
    ((or-exp? exp)
     (let loop ((args (or-exp-args exp)))
       (if (null? args)
           #f
           (or (interp (car args) E A jp)
               (loop (cdr args))))))
    
    ((if-exp? exp)
     (if (interp (if-exp-test exp) E A jp)
         (interp (if-exp-then exp) E A jp)
         (interp (if-exp-else exp) E A jp)))
    
    ((begin-exp? exp)
     (interp-body (begin-exp-exps exp) E A jp))

    ((var-exp? exp)   (env-get (var-exp-name exp) E))
    
    ((set-exp? exp)
     (let ((nval (interp (set-exp-exp exp) E A jp)))
       (env-set! (set-exp-id exp) E nval)))
    
    ;; both let and deploy could be treated as syntactic sugar in the parser
    ;; (let ((x e) ...) b) => ((λ (x ...) b) e ...)
    ;; (deploy ds a e)     => ((deploy-on ds a (λ () e)))
    
    ((let-exp? exp)
     (let ((args   (interp-args (let-exp-args exp) E A jp)))
       (interp-body (let-exp-body exp) (extend-env (let-exp-params exp) args E) A jp)))
    
    
    ((depl-exp? exp)  (let* ((ds  (interp (depl-exp-ds exp) E A jp))
                             (asp (interp (depl-exp-asp exp) E A jp))
                             (dasp (build-dasp ds asp)))
                        (interp-body (depl-exp-body exp)  E (union-cons dasp A) jp)))
    
    ;;---
    
    ((fun-exp? exp)   (let ((E (if (eq? E (get-root-env))  ;; closures should not 
                                   (make-empty-env)        ;; capture root-env
                                   E)))
                        (new-function (fun-exp-params exp)
                                      (fun-exp-body exp)
                                      E A jp)))
    
    ((app-exp? exp)   (let ((cl   (interp (app-exp-fun exp) E A jp))
                            (args (interp-args (app-exp-args exp) E A jp)))
                        (interp-call cl args A jp)))
    
    ((depl-on-exp? exp) (let* ((ds   (interp (depl-on-exp-ds exp) E A jp))
                               (asp  (interp (depl-on-exp-asp exp) E A jp))
                               (dasp (build-dasp ds asp))
                               (f    (interp (depl-on-exp-fun exp) E A jp)))
                          (deploy-on f dasp)
                          'void))
    
    ((this-jp-exp? exp) jp)))


;; create a new function given formals, body and current environments.
(define (new-function formals body E A jp)
  (if (get-target-host) (error 'new-function "new lambda created while serializing (should not happen)"))
  (let ((njp (make-jp 'new #f formals jp (get-current-host) #f)))
    (weave-some A njp)
    (make-closure formals body E (collect-match-d njp A))))
 
;; interpretation of a function call
;; fun and arguments are already values
;; (called from weaver to evaluate pcs/advs)
(define (interp-call f args A jp)
  (check-args args f)
  (let ((njp  (make-jp 'call f args jp (get-current-host) (target-host f)))) ;; create call jp (client-side)
    (weave-some A njp)
    (exec-fun f args (collect-match-c njp A) njp))) ;; on client, collect aspects that propagate on call stack
    
;; evaluate function body in given context
;; closure and arguments are already values
;; (may be called from a skeleton within the transport layer!)
(define (interp-exec cl args A jp)
  (let* ((njp  (make-jp 'exec cl args jp (get-current-host) #f)) ;; create exec jp (server-side)
         (env  (extend-env (closure-params cl) args (closure-env cl)))
         (asps (union A (closure-aspects cl))))
    (weave-some asps njp) 
    (interp-body (closure-body cl) env asps njp))) 

(define (interp-args exps E A jp)
    (map (lambda (x) (interp x E A jp)) exps))

(define (interp-body body E A jp)
  (let loop ((exps body) (val #f))
    (if (null? exps)
        val
        (loop (cdr exps) (interp (car exps) E A jp)))))

;; checks that the number of arguments matches 
;; the number of formal parameters of the function (stub/closure).
;; throws an error if not, returns #t if ok.
(define (check-args args f)
  (let ((count (length args))
        (f-arg-count (arg-count f)))
    (if (eq? f-arg-count count) #t
        (error 'application "bad argument count: expected ~a, given ~a" 
               f-arg-count
               count))))

;;;;;;;;;;;;;;;;;;;;
;; closure/stub API

(define (arg-count f)
  (if (fun-stub? f)
      (fun-stub-arg-count f)
      (length (closure-params f))))

(define (exec-fun f args A jp)
  (if (fun-stub? f)
      (remote-exec f args A jp)
      (interp-exec f args A jp)))

(define (target-host f)
  (if (fun-stub? f)
      (fun-stub-host f) 
      #f))

(define (deploy-on f dasp)
  (if (fun-stub? f) 
      (error 'deploy-on "cannot deploy aspects on a function remotely")
      (set-closure-aspects! f (union-cons dasp (closure-aspects f)))))

;; server.scm defines export-function


;; copying of function

;; list of already-copied functions
;; used during deep copying to avoid cycles
;; (thread local state, reinitialized for each top level copy)
(define _copied (make-thread-cell '()))

;; copies function f (top-level)
(define (copy-function f A jp)
  (thread-cell-set! _copied '())
  (copy-function-l f A jp))

;; recursive copy
(define (copy-function-l f A jp)
  (thread-cell-set! _copied (cons f (thread-cell-ref _copied)))
  (let ((njp (make-jp 'new f (closure-params f) jp (get-current-host) (get-target-host))))
    (weave-some A njp)
    (make-closure (closure-params f)
                  (closure-body f)
                  (env-copy-ctx (closure-env f) A jp)
                  (union (collect-match-d njp A)                       
                         (collect-match-d njp (closure-aspects f)))))) 

;; deeps copies the environment, 
;; making deep copies of lists and using 
;; copy-function of each closures, 
;; using context (A and jp)
(define (env-copy-ctx e A jp)
  (define (copy-value v)
    (cond ((list? v) (map copy-value v))
          ((closure? v) (if (member v (thread-cell-ref _copied)) v
                            (copy-function-l v A jp)))
          (else v)))
  (env-copy e copy-value))

;;;;;;;;;;;;;;;;;;; ASPECTS ;;;;;;;;;;;;;;;

;; builds a deployed aspect structure from the given deployment strategy and aspect
(define (build-dasp ds asp) 
  (make-dasp (car ds)
             (cadr ds)
             (caddr ds)
             (car asp)
             (cadr asp)))

;; two deployed aspects are deemed equal if all their components are equal.
(define ((dasp-equal? d1) d2)
  (and (equal? (dasp-c d1) (dasp-c d2))
       (equal? (dasp-d d1) (dasp-d d2))
       (equal? (dasp-f d1) (dasp-f d2))
       (equal? (dasp-pc d1) (dasp-pc d2))
       (equal? (dasp-adv d1) (dasp-adv d2))))

;; aspect environments
(define (collect-match-c jp asps) (collect-if (lambda (a) (app/prim (dasp-c a) jp)) asps))
(define (collect-match-d jp asps) (collect-if (lambda (a) (app/prim (dasp-d a) jp)) asps))
(define (collect-match-f jp asps) (collect-if (lambda (a) (app/prim (dasp-f a) jp)) asps))

;; union of aspect environments
(define (union ae1 ae2)
  (cond ((empty? ae1) ae2)
        ((memf (dasp-equal? (car ae1)) ae2) (union (cdr ae1) ae2))
        (else (cons (car ae1) (union (cdr ae1) ae2)))))

;; adds an element (cons) only if not present (union semantics)
(define (union-cons da ae)
  (if (memf (dasp-equal? da) ae)
      ae
      (cons da ae)))
         
         
;; weaving
(define (weave-some asps jp)
  (weave-all (collect-match-f jp asps) jp))
   
(define (weave-all asps jp)
  (map (lambda (x) (weave x jp)) asps))

(define (weave asp jp)
  (if (app/prim (dasp-pc asp) jp)
      (app/prim (dasp-adv asp) jp))) 

;; applies the function f, passing jp as argument
;; evaluation is done in an empty aspect environment
(define (app/prim f jp)
  (interp-call f (list jp) '() jp))