;; This prelude contains definitions that are performed when 
;; an interpreter instance is started.
;; In particular, contains typical pointcuts and deployment strategies.

(define *prelude*
  '(begin
     
     ;; jp kind predicates
     (define call?      (fun (jp) (eq? (jp-kind jp) 'call)))
     (define exec?      (fun (jp) (eq? (jp-kind jp) 'exec)))
     (define new?       (fun (jp) (eq? (jp-kind jp) 'new)))
     (define fresh-new? (fun (jp) (and (new? jp) (eq? (jp-fun jp) #f))))
     (define copy?      (fun (jp) (and (new? jp) (not (eq? (jp-fun jp) #f)))))

     ;; remote predicates
     (define remote?      (fun (jp) (not (eq? (jp-target-host jp) #f))))
     (define remote-call? (fun (jp) (and (call? jp) (remote? jp))))
     (define remote-copy? (fun (jp) (and (copy? jp) (remote? jp))))
     
     ;; standard pointcuts 
     (define t (fun (jp) #t))
     (define f (fun (jp) #f))  
     
     (define call (fun (f) 
                       (fun (jp) (and (call? jp)
                                      (eq? (jp-fun jp) f)))))
     
     (define exec (fun (f) 
                       (fun (jp) (and (exec? jp)
                                      (eq? (jp-fun jp) f)))))
     
     (define within (fun (f) 
                         (fun (jp) (eq? (and (jp-parent jp)
                                             (jp-fun (jp-parent jp)) f)))))
     
     (define cflow (fun (pc) 
                        (fun (jp) (or (pc jp)
                                      ((cflowbelow pc) jp)))))
     
     (define cflowbelow (fun (pc) 
                             (fun (jp) (and (jp-parent jp)
                                            ((cflow pc) (jp-parent jp))))))
     
     (define && (fun (pc1 pc2)
                     (fun (jp) (and (pc1 jp) (pc2 jp)))))
     
     (define || (fun (pc1 pc2)
                     (fun (jp) (or (pc1 jp) (pc2 jp)))))
     
     (define ! (fun (pc) 
                    (fun (jp) (not (pc jp)))))
     
     ;; local only versions of cflow
     (define local-cflow 
       (fun (pc)
            (fun (jp) (or (pc jp)
                          (and (jp-parent jp)
                               (if (eq? (jp-host (jp-parent jp)) 
                                        (jp-host jp))
                                   ((local-cflow pc) (jp-parent jp))
                                   #f))))))
     
     (define local-cflowbelow 
       (fun (pc)
            (fun (jp) (and (jp-parent jp)
                           (if (eq? (jp-host (jp-parent jp)) 
                                    (jp-host jp))
                               ((local-cflow pc) (jp-parent jp))
                               #f)))))
     
     ;; host/port utils
     (define jp-hostname (fun (jp) (host-name (jp-host jp))))
     (define jp-port     (fun (jp) (host-port (jp-host jp))))
     (define in-port     (fun (port) (fun (jp) (eq? (jp-port jp) port))))
     ;...etc...
     
     ;; deployment strategies: constructor
     ; (deployment strategies are represented as lists)
     (define make-ds (fun (c d f) (list c d f)))

     ; accessors 
     (define ds-c (fun (ds) (car ds)))
     (define ds-d (fun (ds) (cadr ds)))
     (define ds-f (fun (ds) (caddr ds)))
     
     ;; standard deployment strategies "a la AspectScheme"
     (define as-static  (make-ds f t t))
     (define as-dynamic (make-ds t f t))
     ;; propagate in both directions
     (define ds-all (make-ds t t t))

     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
     ;; distributed deployment strategies ----
     
     ;; curried version of eq? (needed to define later functions)
     (define eqc?
       (fun (v1)
            (fun (v2)
                 (eq? v1 v2))))
     
     ;; global propagation and activation is the default
     (define propagate-global (fun (ds) ds)) 
     (define active-global (fun (ds) ds))
     
     ;; transforms a deployment strategy to control
     ;; propagation: only to hosts matched by predicate hp
     (define propagate-if 
       (fun (hp)
            (fun (ds)
                 (make-ds ((restrict-target hp) (ds-c ds))
                          ((restrict-target hp) (ds-d ds))
                          (ds-f ds)))))
     
     ;; transforms a deployment strategy to control
     ;; activation: only on hosts matched by predicate hp
     (define active-if
       (fun (hp)
            (fun (ds) 
                 (make-ds (ds-c ds)
                          (ds-d ds)
                          ((restrict-current hp) (ds-f ds))))))
     
     ;; restricts a pointcut based on a predicate (hp) on the current host
     (define restrict-current
       (fun (hp)
            (fun (pc)
                 (fun (jp)
                      (and (hp (jp-host jp))
                           (pc jp))))))
     
     ;; restricts a pointcut based on a predicate (hp) on the target host
     (define restrict-target
       (fun (hp)
            (fun (pc)
                 (fun (jp)
                      (and (remote? jp)
                           (hp (jp-target-host jp))
                           (pc jp))))))
     
     ;; host predicates for current/remote hosts     
     (define is-current
       (eqc? current-host))
     
     (define is-remote
       (fun (h) (not (is-current h))))
          
     ;; standard local/remote distributed strategies
     (define propagate-local
       (propagate-if is-current))
     
     (define active-local
       (active-if is-current))
     
     (define propagate-remote
       (propagate-if is-remote))
     
     (define active-remote
       (active-if is-remote))

     
     ))
     